/*
 * Implementation of Hungarian method (Harold Kuhn, 1955)
 * Copyright (C) 2009 Zhi Yan <yz@ai.univ-paris8.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <algorithm>
#include <cstdlib>
#include "hungarian.hpp"

void debugPrint(int step, double **probmat, int n) {
  int i, j;
  
  std::cout << "\n--- hungarianStep" << step << "() ---\n";
  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      std::cout << probmat[i][j] << "\t";
    }
    std::endl(std::cout);
  }
}

/* 
 * Step 1: Subtract the elements of each row by the row minimum,
 *         subtract the elements of each column by the column minimum.
 *
 * @param probmat: problem matrix
 * @param n: n×n matrix
 */
void hungarianStep1(double **probmat, int n) {
  double row_min, col_min;
  int i, j;
  
  // row
  for(i = 0; i < n; i++) {
    row_min = probmat[i][0];
    for(j = 1; j < n; j++)
      row_min = std::min(probmat[i][j], row_min);
    for(j = 0; j < n; j++)
      probmat[i][j] -= row_min;
  }
  
  // column
  for(j = 0; j < n; j++) {
    col_min = probmat[0][j];
    for(i = 1; i < n; i++)
      col_min = std::min(probmat[i][j], col_min);
    for(i = 0; i < n; i++)
      probmat[i][j] -= col_min;
  }
  
  // debugPrint(1, probmat, n);
}

/* Step 2: A test for optimality: assignment-list-size < n ?
 *
 * @param probmat: problem matrix
 * @param n: n×n matrix
 * @param asglist: assignment list
 */
bool compare(Zero a, Zero b) {return (a.getAmount() < b.getAmount());}
bool hungarianStep2(double **probmat, int n, std::list<Element> &asglist) {
  bool tag[n][n];
  int i, j;
  
  std::list<Zero> zero_list;
  std::list<Zero>::iterator zero_it;
  
  // tag initialisation
  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      tag[i][j] = false;
  
  // Find the number of zero in each row.
  int zero_amount;
  asglist.clear();
  for(i = 0; i < n; i++) {
    zero_amount = 0;
    for(j = 0; j < n; j++) {
      if(probmat[i][j] == 0)
	zero_amount++;
    }
    if(zero_amount > 0) {
      Zero zero(i, zero_amount);
      zero_list.push_back(zero);
    }
  }
  zero_list.sort(compare);
  
  // debug print
  // for(zero_it = zero_list.begin(); zero_it != zero_list.end(); zero_it++)
  //   std::cout << zero_it->getRow();
  // std::cout << std::endl;
  
  // Find all possible assignments.
  while((int)zero_list.size() > 0) {    
    Zero zero = zero_list.front();
    
    if(zero.getAmount() > 0) {
      for(j = 0; j < n; j++) {
	if(probmat[zero.getRow()][j] == 0 && !tag[zero.getRow()][j]) {
	  Element element(zero.getRow(), j);
	  asglist.push_back(element);
	  zero_list.pop_front();
	  
	  for(i = 0; i < n; i++) {
	    if(probmat[i][j] == 0) {
	      tag[i][j] = true;
	      for(zero_it = zero_list.begin(); zero_it != zero_list.end(); zero_it++)
	      	if(zero_it->getRow() == i)
		  zero_it->setAmount(zero_it->getAmount()-1);
	    }
	  }
	  break;
	}
      }
    } else {
      zero_list.pop_front();
    }
    zero_list.sort(compare);
  }
  
  zero_list.clear();
  zero_list.~list();
  
  return ((int)asglist.size() == n);
}

/*
 * Step 3: First, draw lines across rows and columns in such a way that all zeros
 *         are covered and that the minimun number of lines have been used.
 *         Then, find the smallest element which is not covered by any of the lines.
 *         Finally, subtract the smallest element from each element which is not covered by the lines
 *         and add it to each element which is covered by a horizontal and a vertical line.
 * 
 * @param probmat: problem matrix
 * @param n: n×n matrix
 */
void hungarianStep3(double **probmat, int n) {
  bool tag[n][n];
  int i, j, ii, jj;
  
  // tag initialisation
  for(i = 0; i < n; i++)
    for(j = 0; j < n; j++)
      tag[i][j] = false;
  
  // Draw lines across rows and columns in such a way that all zeros
  // are covered and that the minimun number of lines have been used.
  int n0inrow, n0incol;
  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      if(probmat[i][j] == 0 && !tag[i][j]) {
	n0inrow = n0incol = 1;
	
	for(jj = 0; jj < n; jj++) {
	  if(probmat[i][jj] == 0 && jj != j)
	    n0inrow++;
	}
	for(ii = 0; ii < n; ii++) {
	  if(probmat[ii][j] == 0 && ii != i)
	    n0incol++;
	}
	
	if(n0inrow > n0incol) {
	  for(jj = 0; jj < n; jj++)
	    tag[i][jj] = true;
	} else {
	  for(ii = 0; ii < n; ii++)
	    tag[ii][j] = true;
	}
      }
    }
  }
  
  // debug print
  // std::cout << "\n--- draw lines ---\n";
  // for(i = 0; i < n; i++) {
  //   for(j = 0; j < n; j++) {
  //     if(tag[i][j]) std::cout << "x\t";
  //     else          std::cout << ".\t";
  //   }
  //   std::endl(std::cout);
  // }
  
  // Find the smallest element which is not covered by any of the lines.
  double smallest = RAND_MAX;
  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      if(!tag[i][j])
	smallest = std::min(probmat[i][j], smallest);
    }
  }
  
  // Subtract the smallest element from each element which is not covered by the lines
  // and add it to each element which is covered by a horizontal and a vertical line.
  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      if(!tag[i][j]) {
  	for(jj = 0; jj < n; jj++)
  	  probmat[i][jj] -= smallest;
  	break;
      }
    }
  }
  for(i = 0; i < n; i++) {
    for(j = 0; j < n; j++) {
      if(probmat[i][j] < 0) {
  	for(ii = 0; ii < n; ii++)
  	  probmat[ii][j] += smallest;
  	break;
      }
    }
  }
  
  // debugPrint(3, probmat, n);
}
