/*
 * Implementation of Hungarian method (Harold Kuhn, 1955)
 * Copyright (C) 2009 Zhi Yan <yz@ai.univ-paris8.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _HUNGARIAN_HPP
#define _HUNGARIAN_HPP

#include <list>

// matrix element class
class Element {
private:
  int row;
  int column;
  
public:
  Element();
  Element(int _row, int _column) {
    row = _row;
    column = _column;
  }
  ~Element(){}
  
  int getRow() {return row;}
  int getColumn() {return column;}
};

class Zero {
private:
  int row;
  int amount;
  
public:
  Zero();
  Zero(int _row, int _amount) {
    row = _row;
    amount = _amount;
  }
  ~Zero(){}
  
  void setAmount(int _amount) {
    amount = _amount;
  }
  
  int getRow() {return row;}
  int getAmount() {return amount;}
};

/*
 * Step 1: Subtract the elements of each row by the row minimum,
 *         subtract the elements of each column by the column minimum.
 *
 * @param probmat: problem matrix
 * @param n: n×n matrix
 */
void hungarianStep1(double **probmat, int n);

/* Step 2: A test for optimality: assignment-list-size < n ?
 *
 * @param probmat: problem matrix
 * @param n: n×n matrix
 * @param asglist: assignment list
 */
bool hungarianStep2(double **probmat, int n, std::list<Element> &asglist);

/*
 * Step 3: First, draw lines across rows and columns in such a way that all zeros
 *         are covered and that the minimun number of lines have been used.
 *         Then, find the smallest element which is not covered by any of the lines.
 *         Finally, subtract the smallest element from each element which is not covered by the lines
 *         and add it to each element which is covered by a horizontal and a vertical line.
 * 
 * @param probmat: problem matrix
 * @param n: n×n matrix
 */
void hungarianStep3(double **probmat, int n);

#endif /* !_HUNGARIAN_HPP */
